/****************************************************************************************/
/* Basis.ino                                                                            */
/* Collection of basic routines                                                         */
/*                                                                                      */
/* BUTTON HANDLING:                                                                     */
/* SUB_LONG_PRESS     Subroutine to detect and classify button pressings                */
/* SUB_TOGGLE_MODE    Subroutine to toggle mode activated/deactivated speech recognition*/
/* SUB_TOGGLE_LOAD    Subroutine to toggle between activated/deactivated load mode      */
/* SUB_TOGGLE_RECOVER Subroutine to toggle between activated/deactivated recover mode   */
/* SUB_RESET_VIC      Subroutine to reset vic Control stamp                             */
/* SUB_ESP_BOOTLOADER Subroutine to enter ESP32 bootloader                              */
/*                                                                                      */
/* HOMEMATIC ROUTINES:                                                                  */
/* SUB_TRANS_HOMEMATIC Subroutine to tranmit byte value via homematic interface         */
/*                                                                                      */
/* OTHER ROUTINES:                                                                      */
/* SUB_ERROR_ROUTINE  Error handler with endless red blinking                           */
/****************************************************************************************/


/****************************************************************************************/
/****************************************************************************************/
/* SUBROUTINES FOR BUTTON HANDLING                                                      */
/****************************************************************************************/ 
/****************************************************************************************/


/****************************************************************************************/
/* Subroutine to detect and classify button pressings                                   */
/****************************************************************************************/ 
void SUB_LONG_PRESS (Button2& btn) 
{  
  unsigned int time = btn.wasPressedFor();        /* Time in ms how long button was pressed */

  /* Button A power/mode */
  if (btn == BUTTON_A)                            /* Button A for power/mode pressed ? */
  {
    if (time >= 2000)                             /* Longpress detected ? */
    {
      #ifdef _DEBUG_      
        Serial.println("Button A: Power button long pressed = Power off");
      # endif        
      BUTTON_LONG_POWER_PRESSED = true;           /* Set the power button long pressed flag */ 
    }
    else if ((time > 100)&& (time < 2000))        /* Shortpress detected ? */
    {
      #ifdef _DEBUG_      
        Serial.println("Button A: Power button short pressed = Toggle mode");
      # endif         
      BUTTON_SHORT_POWER_PRESSED = true;          /* Set the power button short pressed flag */     
    }
  } 

  /* Button B load/recover */
  if (btn == BUTTON_B)                            /* Button B for load/recover pressed ? */
  {
    if (time >= 2000)                             /* Longpress detected ? */
    {
      #ifdef _DEBUG_      
        Serial.println("Button B: Load/Recover button long pressed = Recover mode");
      # endif        
      BUTTON_RECOVER_PRESSED = true;              /* Set the recover button pressed flag */ 
    }
    else if ((time > 100)&& (time < 2000))        /* Shortpress detected ? */
    {
      #ifdef _DEBUG_      
        Serial.println("Button B: Load/Recover button short pressed = Load mode");
      # endif  
      BUTTON_LOAD_PRESSED = true;                 /* Set the load button pressed flag */ 
    }
  }
  
  /* Button C boot/reset */
  if (btn == BUTTON_C)                            /* Button C for boot/reset pressed ? */
  {
    if (time >= 2000)                             /* Longpress detected ? */
    {
      #ifdef _DEBUG_      
        Serial.println("Button C: Boot/Reset button long pressed = ESP32 Bootloader mode");
      # endif        
      BUTTON_BOOTLOADER_PRESSED = true;           /* Set the ESP32 boatloader button pressed flag */
    }
    else if ((time > 100)&& (time < 2000))        /* Shortpress detected ? */
    {
      #ifdef _DEBUG_      
        Serial.println("Button C: Boot/Reset button short pressed = Reset vic Control go");
      # endif  
      BUTTON_RESET_PRESSED = true;                /* Set the vic Control stamp reset button pressed flag */ 
    }
  }
}


/****************************************************************************************/
/* Subroutine to change the operation mode: Activated/Deactivated speech recognition    */
/****************************************************************************************/ 
void SUB_TOGGLE_MODE (void)
{
  if (STATE_POWER_ON == false)                    /* Mode = Speech recognition deactivated ? */
  {
    /* Empty receiver buffer */
    while (Serial1.available())                   /* Not empty receiver buffer? */
    {
      Serial1.read();                             /* Read buffer */
    }
    #ifdef _DEBUG_
      Serial.println ("  Activate speech recognizer...");
    # endif    
    Serial1.print (SPEECH_ACTIVATE);              /* Request to start speech recognition */ 
    delay (MESSAGE_RECEIVE_DELAY);                /* Delay to receive answer from vic Control stamp */
    if (SUB_RECEIVE_ANSWER_VIC (4)== false)       /* Receipt of no error (Mode = 4) not OK ? */
    {
      SUB_ERROR_ROUTINE ();                       /* Call sub routine for error handling in endless red blinking loop */
    } 
    SUB_NEO_SEND_COLOR(128, 0, 0);                /* Set green color on Neopixel LED */       
    STATE_POWER_ON = true;                        /* Set power state to On */  
    digitalWrite(MUTE, HIGH);                     /* Set Mute to high = Unmuted */    
    #ifdef _DEBUG_
      Serial.println ("  Playing ready message...");
    # endif
    Serial1.print (PROMPT_4);                     /* Play ready prompt */ 
    delay (MESSAGE_RECEIVE_DELAY);                /* Delay to receive answer from vic Control stamp */
    if (SUB_RECEIVE_ANSWER_VIC (4)== false)       /* Receipt of no error (Mode = 4) not OK ? */
    {
      SUB_ERROR_ROUTINE ();                       /* Call sub routine for error handling in endless red blinking loop */
    }
    #ifdef _DEBUG_
      Serial.println ();
    # endif    
  }
  else                                            /* Mode = Speech recognition activated ? */
  {     
    /* Empty receiver buffer */
    while (Serial1.available())                   /* Not empty receiver buffer? */
    {
      Serial1.read();                             /* Read buffer */
    }   
    #ifdef _DEBUG_
      Serial.println ("  Deactivate speech recognizer ...");
    # endif       
    Serial1.print (SPEECH_DEACTIVATE);            /* Request to stop speech recognition */ 
    delay (MESSAGE_RECEIVE_DELAY);                /* Delay to receive answer from vic Control stamp */
    if (SUB_RECEIVE_ANSWER_VIC (4)== false)       /* Receipt of no error (Mode = 4)not OK ? */
    {
        SUB_ERROR_ROUTINE ();                     /* Call sub routine for error handling in endless red blinking loop */
    }      
    #ifdef _DEBUG_
      Serial.println ("  Playing standby message...");
    # endif
    SUB_NEO_SEND_COLOR(36, 128, 0);               /* Set orange color on Neopixel LED */  
    STATE_POWER_ON = false;                       /* Set power state to On */     
    Serial1.print (PROMPT_5);                     /* Play standby prompt */ 
    delay (MESSAGE_RECEIVE_DELAY);                /* Delay to receive answer from vic Control stamp */
    if (SUB_RECEIVE_ANSWER_VIC (4)== false)       /* Receipt of no error (Mode = 4) not OK ? */
    {
      SUB_ERROR_ROUTINE ();                       /* Call sub routine for error handling in endless red blinking loop */
    }
    delay (PLAY_TIME_DELAY);                      /* Delay until message played */
    digitalWrite(MUTE, LOW);                      /* Set Mute to low = Muted */ 
    #ifdef _DEBUG_
      Serial.println ();
    # endif      
  }
  BUTTON_SHORT_POWER_PRESSED = false;             /* Reset the short pressed flag */
}


/****************************************************************************************/
/* Subroutine to toggle between activated/deactivated load mode                         */
/****************************************************************************************/
void  SUB_TOGGLE_LOAD (void)
{
  if (STATE_LOAD_ON == false)                     /* Normal mode ? */
  {
    STATE_LOAD_ON = true;                         /* Set load mode to On */  
    SUB_NEO_SEND_COLOR(0, 0, 128);                /* Set blue color on Neopixel LED */ 
    #ifdef _DEBUG_
      Serial.println ("  Activate load mode: Done");
      Serial.println ("  Playing load message...");
    # endif
    digitalWrite(MUTE, HIGH);                     /* Set Mute to high = Unmuted */      
    Serial1.print (PROMPT_6);                     /* Play load prompt */ 
    delay (MESSAGE_RECEIVE_DELAY);                /* Delay to receive answer from vic Control stamp */
    if (SUB_RECEIVE_ANSWER_VIC (4)== false)       /* Receipt of no error (Mode = 4) not OK ? */
    {
      SUB_ERROR_ROUTINE ();                       /* Call sub routine for error handling in endless red blinking loop */
    }
    digitalWrite(MULTIPLEXER, HIGH);              /* Set vic Control stamp by USB */
    delay (MULTIPLEXER_DELAY);                    /* Multiplexer setup delay */
    /* Empty receiver buffer */
    while (Serial1.available())                   /* Not empty receiver buffer? */
    {
      Serial1.read();                             /* Read buffer */
    }  
    if (STATE_POWER_ON == false)                  /* Power state is Off ? */
    {  
      delay (PLAY_TIME_DELAY);                    /* Delay until message played */
      digitalWrite(MUTE, LOW);                    /* Set Mute to low = Muted */ 
    }
  }
  else                                            /* Load mode ? */
  {    
    STATE_LOAD_ON = false;                        /* Set load state to Normal */  
    #ifdef _DEBUG_
      Serial.println ("  Deactivate load mode: Done");
    # endif  
    digitalWrite(MULTIPLEXER, LOW);               /* Set vic Control stamp by ESP */    
    delay (MULTIPLEXER_DELAY);                    /* Multiplexer setup delay */
    /* Empty receiver buffer */
    while (Serial1.available())                   /* Not empty receiver buffer? */
    {
      Serial1.read();                             /* Read buffer */
    }      
    digitalWrite(MUTE, HIGH);                     /* Set Mute to high = Unmuted */
    if (STATE_POWER_ON == true)                   /* Power state is On ? */
    {
      #ifdef _DEBUG_
        Serial.println ("  Playing ready message...");
      # endif
      Serial1.print (PROMPT_4);                   /* Play ready prompt */ 
      delay (MESSAGE_RECEIVE_DELAY);              /* Delay to receive answer from vic Control stamp */
      if (SUB_RECEIVE_ANSWER_VIC (4)== false)     /* Receipt of no error (Mode = 4) not OK ? */
      {
        SUB_ERROR_ROUTINE ();                     /* Call sub routine for error handling in endless red blinking loop */
      }      
      SUB_NEO_SEND_COLOR(128, 0, 0);              /* Set green color on Neopixel LED */  
    }
    else                                          /* Power on is Off ? */
    {
      #ifdef _DEBUG_
        Serial.println ("  Playing standby message...");
      # endif
      Serial1.print (PROMPT_5);                   /* Play standby prompt */ 
      delay (MESSAGE_RECEIVE_DELAY);              /* Delay to receive answer from vic Control stamp */
      if (SUB_RECEIVE_ANSWER_VIC (4)== false)     /* Receipt of no error (Mode = 4) not OK ? */
      {
        SUB_ERROR_ROUTINE ();                     /* Call sub routine for error handling in endless red blinking loop */
      }            
      SUB_NEO_SEND_COLOR(36, 128, 0);             /* Set orange color on Neopixel LED */   
      delay (PLAY_TIME_DELAY);                    /* Delay until message played */
      digitalWrite(MUTE, LOW);                    /* Set Mute to low = Muted */               
    }       
  }  
  #ifdef _DEBUG_    
    Serial.println (); 
  # endif   
  BUTTON_LOAD_PRESSED = false;                    /* Reset the load button pressed flag */
}


/****************************************************************************************/
/* Subroutine to toggle between activated/deactivated recover mode                      */
/****************************************************************************************/
void  SUB_TOGGLE_RECOVER (void)
{
  if (STATE_RECOVER_ON == false)                  /* Normal mode ? */
  {
    #ifdef _DEBUG_
      Serial.println ("  Activate recover mode...");
    # endif       
    SUB_NEO_SEND_COLOR(0, 128, 64);               /* Set purple color on Neopixel LED */ 
    #ifdef _DEBUG_
      Serial.println ("  Playing recover message...");
    # endif
    digitalWrite(MUTE, HIGH);                     /* Set Mute to high = Unmuted */    
    Serial1.print (PROMPT_7);                     /* Play recover prompt */ 
    delay (MESSAGE_RECEIVE_DELAY);                /* Delay to receive answer from vic Control stamp */
    if (SUB_RECEIVE_ANSWER_VIC (4)== false)       /* Receipt of no error (Mode = 4) not OK ? */
    {
      SUB_ERROR_ROUTINE ();                       /* Call sub routine for error handling in endless red blinking loop */
    }      
    digitalWrite(MULTIPLEXER, HIGH);              /* Set vic Control stamp by USB */
    digitalWrite(RECOV_VICCONTROL, LOW);          /* Activate recovery line */
    delay (LONG_BLINK_DELAY);                     /* Blink delay for LED long blinking */
    /* Empty receiver buffer */
    while (Serial1.available())                   /* Not empty receiver buffer? */
    {
      Serial1.read();                             /* Read buffer */
    }       
    SUB_NEO_SEND_COLOR(0, 0, 0);                  /* Switch Neopixel LED off */
    delay (LONG_BLINK_DELAY);                     /* Blink delay for LED long blinking */
    digitalWrite(MUTE, LOW);                      /* Set Mute to low = Muted */
    SUB_NEO_SEND_COLOR(0, 128, 64);               /* Set purple color on Neopixel LED */
    digitalWrite(RESET_VICCONTROL, LOW);          /* Activate Reset line */
    delay (LONG_BLINK_DELAY);                     /* Blink delay for LED long blinking */
    SUB_NEO_SEND_COLOR(0, 0, 0);                  /* Switch Neopixel LED off */
    delay (LONG_BLINK_DELAY);                     /* Blink delay for LED long blinking */
    SUB_NEO_SEND_COLOR(0, 128, 64);               /* Set purple color on Neopixel LED */
    digitalWrite(RESET_VICCONTROL, HIGH);         /* Deactivate Reset line */
    delay (LONG_BLINK_DELAY);                     /* Blink delay for LED long blinking */
    SUB_NEO_SEND_COLOR(0, 0, 0);                  /* Switch Neopixel LED off */
    delay (LONG_BLINK_DELAY);                     /* Blink delay for LED long blinking */
    SUB_NEO_SEND_COLOR(0, 128, 64);               /* Set purple color on Neopixel LED */
    delay (LONG_BLINK_DELAY);                     /* Blink delay for LED long blinking */
    SUB_NEO_SEND_COLOR(0, 0, 0);                  /* Switch Neopixel LED off */
    delay (LONG_BLINK_DELAY);                     /* Blink delay for LED long blinking */
    SUB_NEO_SEND_COLOR(0, 128, 64);               /* Set purple color on Neopixel LED */
    digitalWrite(RECOV_VICCONTROL, HIGH);         /* Deactivate recovery line */
    STATE_RECOVER_ON = true;                      /* Set recover mode to On */             
    #ifdef _DEBUG_
      Serial.println ("  Recover mode activated!");
    # endif 
               
  }
  else                                            /* Recover mode ? */
  {  
    #ifdef _DEBUG_
      Serial.println ("  Deactivate recover mode by reseting device ...");
    # endif    
    SUB_NEO_SEND_COLOR(0, 0, 0);                  /* Switch Neopixel LED off */ 
    digitalWrite(MULTIPLEXER, LOW);               /* Set vic Control stamp by ESP */ 
    digitalWrite(RESET_VICCONTROL, LOW);          /* Reset vic Control stamp */                 
    delay (LONG_BLINK_DELAY);                     /* Blink delay for LED long blinking */
    /* Empty receiver buffer */
    while (Serial1.available())                   /* Not empty receiver buffer? */
    {
      Serial1.read();                             /* Read buffer */
    }       
    digitalWrite(RESET_VICCONTROL, HIGH);         /* Remove Reset vic Control stamp */ 
    SUB_NEO_SEND_COLOR(0, 128, 64);               /* Set purple color on Neopixel LED */
    delay (LONG_BLINK_DELAY);                     /* Blink delay for LED long blinking */  
    LOOP_COUNTER = 0;                             /* Reset the loop counter */
    do                                            /* Repeat until boot message received or timeout */
    {
      SUB_NEO_SEND_COLOR(0, 0, 0);                /* Switch Neopixel LED off */
      delay (LONG_BLINK_DELAY);                   /* Blink delay for LED long blinking */
      SUB_NEO_SEND_COLOR(0, 128, 64);             /* Set purple color on Neopixel LED */
      delay (LONG_BLINK_DELAY);                   /* Blink delay for LED long blinking */
      if (Serial1.available() > 0)                /* Received something from vic Control stamp ? */
      {
        if (SUB_RECEIVE_ANSWER_VIC (0)== true)    /* Receipt of boot message (Mode 0) OK ? */
        {
          LOOP_COUNTER = 255;                     /* Set for break loop */         
        }
        else                                      /* Receipt of boot message (Mode 0) not OK ? */
        {
          SUB_ERROR_ROUTINE ();                   /* Call sub routine for error handling in endless red blinking loop */
        }
      }
      else                                        /* Nothing received so far ? */
      {
        LOOP_COUNTER += 1;                        /* Increase loop counter */    
      }
    }
    while (LOOP_COUNTER < 10);                    /* Wait for maximum of 20 seconds */

    if (LOOP_COUNTER == 10)                       /* Timeout */
    {
      #ifdef _DEBUG_
        Serial.println ("Boot timeout error!");
      # endif  
      SUB_ERROR_ROUTINE ();                       /* Call sub routine for error handling in endless red blinking loop */       
    } 
    SUB_NEO_SEND_COLOR(128, 0, 0);                /* Set green color on Neopixel LED */       
    STATE_POWER_ON = true;                        /* Set power state to On */ 
    digitalWrite(MUTE, HIGH);                     /* Set Mute to high = Unmuted */
    STATE_RECOVER_ON = false;                     /* Set recover mode to normal mode */ 
    #ifdef _DEBUG_
      Serial.println ("  Recover mode deactivated!");
      Serial.println ("  Playing ready message...");
    # endif
    digitalWrite(MUTE, HIGH);                     /* Set Mute to high = Unmuted */
    Serial1.print (PROMPT_4);                     /* Play ready prompt */ 
    delay (MESSAGE_RECEIVE_DELAY);                /* Delay to receive answer from vic Control stamp */
    if (SUB_RECEIVE_ANSWER_VIC (4)== false)       /* Receipt of no error (Mode = 4) not OK ? */
    {
      SUB_ERROR_ROUTINE ();                       /* Call sub routine for error handling in endless red blinking loop */
    }     
  } 
  #ifdef _DEBUG_    
    Serial.println (); 
  # endif  
  BUTTON_RECOVER_PRESSED = false;                 /* Reset the load button pressed flag */
}


/****************************************************************************************/
/* Subroutine to reset vic Control stamp                                                */
/****************************************************************************************/
void SUB_RESET_VIC (void)
{
  #ifdef _DEBUG_
    Serial.print ("  Reseting vic Control stamp... ");
  # endif   
  SUB_NEO_SEND_COLOR(0, 0, 0);                    /* Switch Neopixel LED off */
  digitalWrite(MUTE, LOW);                        /* Set Mute to low = Muted */ 
  digitalWrite(MULTIPLEXER, LOW);                 /* Set vic Control stamp by ESP */ 
  digitalWrite(RESET_VICCONTROL, LOW);            /* Reset vic Control stamp */                 
  delay (LONG_BLINK_DELAY);                       /* Blink delay for LED long blinking */
  /* Empty receiver buffer */
  while (Serial1.available())                     /* Not empty receiver buffer? */
  {
    Serial1.read();                               /* Read buffer */
  } 
  digitalWrite(RESET_VICCONTROL, HIGH);           /* Remove Reset vic Control stamp */ 
  SUB_NEO_SEND_COLOR(0, 128, 0);                  /* Set red color on Neopixel LED */
  delay (LONG_BLINK_DELAY);                       /* Blink delay for LED long blinking */  
  LOOP_COUNTER = 0;                               /* Reset the loop counter */
  do                                              /* Repeat until boot message received or timeout */
  {
    SUB_NEO_SEND_COLOR(0, 0, 0);                  /* Switch Neopixel LED off */
    delay (LONG_BLINK_DELAY);                     /* Blink delay for LED long blinking */ 
    SUB_NEO_SEND_COLOR(0, 128, 0);                /* Set red color on Neopixel LED */
    delay (LONG_BLINK_DELAY);                     /* Blink delay for LED long blinking */
    if (Serial1.available() > 0)                  /* Received something from vic Control stamp ? */
    {
      if (SUB_RECEIVE_ANSWER_VIC (0)== true)      /* Receipt of boot message (Mode 0) OK ? */
      {
        LOOP_COUNTER = 255;                       /* Set for break loop */         
      }
      else                                        /* Receipt of boot message (Mode 0) not OK ? */
      {
        SUB_ERROR_ROUTINE ();                     /* Call sub routine for error handling in endless red blinking loop */
      }
    }
    else                                          /* Nothing received so far ? */
    {
      LOOP_COUNTER += 1;                          /* Increase loop counter */    
    }
  }
  while (LOOP_COUNTER < 10);                      /* Wait for maximum of 20 seconds */

  if (LOOP_COUNTER == 10)                         /* Timeout */
  {
    #ifdef _DEBUG_
      Serial.println ("Boot timeout error!");
    # endif  
    SUB_ERROR_ROUTINE ();                         /* Call sub routine for error handling in endless red blinking loop */       
  } 
  SUB_NEO_SEND_COLOR(128, 0, 0);                  /* Set green color on Neopixel LED */       
  STATE_POWER_ON = true;                          /* Set power state to On */ 
  STATE_LOAD_ON = false;                          /* Set load state to Normal */
  STATE_RECOVER_ON = false;                       /* Set recover state to Normal */  
  digitalWrite(MUTE, HIGH);                       /* Set Mute to high = Unmuted */  
  #ifdef _DEBUG_
    Serial.println ("  Playing ready message...");
  # endif
  Serial1.print (PROMPT_4);                       /* Play ready prompt */ 
  delay (MESSAGE_RECEIVE_DELAY);                  /* Delay to receive answer from vic Control stamp */
  if (SUB_RECEIVE_ANSWER_VIC (4)== false)         /* Receipt of no error (Mode = 4) not OK ? */
  {
    SUB_ERROR_ROUTINE ();                         /* Call sub routine for error handling in endless red blinking loop */
  } 
  #ifdef _DEBUG_
    Serial.println ();
  # endif   
  BUTTON_RESET_PRESSED = false;                   /* Reset the reset button pressed flag */       
} 


/****************************************************************************************/
/* Subroutine to enter ESP32 bootloader                                                 */
/****************************************************************************************/
void SUB_ESP_BOOTLOADER (void)
{
  byte KEY_VALUE;                                 /* Key Value of pressed button */
  unsigned long AWAKE_END;                        /* Variable for determing display timeout */
  unsigned long CURRENT_TIME;                     /* Variable for determing display timeout */
  boolean BLINK_FLAG;                             /* True: Led is on, false = LED is off */

  SUB_NEO_SEND_COLOR(128, 0, 128);                /* Set cyan color on Neopixel LED */
  #ifdef _DEBUG_
    Serial.println ("  Initializing bootloader...");
    Serial.println ("  Playing bootloader message...");
  # endif
  digitalWrite(MUTE, HIGH);                       /* Set Mute to high = Unmuted */  
  Serial1.print (PROMPT_8);                       /* Play bootloader prompt */  
  delay (MESSAGE_RECEIVE_DELAY);                  /* Delay to receive answer from vic Control stamp */
  if (SUB_RECEIVE_ANSWER_VIC (4)== false)         /* Receipt of no error (Mode = 4) not OK ? */
  {
    SUB_ERROR_ROUTINE ();                         /* Call sub routine for error handling in endless red blinking loop */
  } 
  delay (PLAY_TIME_DELAY);                        /* Delay until message played */
  digitalWrite(MUTE, LOW);                        /* Set Mute to low = Muted */    
    
  SUB_WLAN_CONNECT ();                            /* Call subroutine to connect fo Wifi */

  #ifdef _DEBUG_
      Serial.println ("  Bootloader activated!");
  # endif
  
  ArduinoOTA
    .onStart([]()                                 /* On start */
    {
      String type;
      if (ArduinoOTA.getCommand() == U_FLASH)     /* Sketch upload ? */
        type = "sketch";
      else                                        /* U_SPIFFS selected ? */
        type = "filesystem";
      /* NOTE: if updating SPIFFS this would be the place to unmount SPIFFS using SPIFFS.end() */
      Serial.println("  Start updating " + type);
    })
    
    .onEnd([]()                                   /* Finished upload */ 
    {
      Serial.println("\nEnd");
    })

    /* On progress */
    .onProgress([](unsigned int progress, unsigned int total) 
    {
      Serial.printf("  Progress: %u%%\r", (progress / (total / 100)));
    })
    
    .onError([](ota_error_t error)                /* On error */
    {
      Serial.printf("  Error[%u]: ", error);
      if (error == OTA_AUTH_ERROR) Serial.println("Auth Failed");
      else if (error == OTA_BEGIN_ERROR) Serial.println("Begin Failed");
      else if (error == OTA_CONNECT_ERROR) Serial.println("Connect Failed");
      else if (error == OTA_RECEIVE_ERROR) Serial.println("Receive Failed");
      else if (error == OTA_END_ERROR) Serial.println("End Failed");
    });

  ArduinoOTA.begin();                             /* Start OTA */

  /* Endless loop to wait for request from Arduino IDE */
  CURRENT_TIME = millis();                        /* Get actual time value */
  AWAKE_END = CURRENT_TIME + LONG_BLINK_DELAY;    /* Set new blink value to long blink delay */
  SUB_NEO_SEND_COLOR(128, 0, 128);                /* Set cyan color on Neopixel LED */  
  BLINK_FLAG = true;                              /* Set blink flag true */
  do
  {
    ArduinoOTA.handle();                          /* OTA handler */
    /* Check if display should be switched off after timeout */
    CURRENT_TIME = millis();                      /* Get actual time value */
    /* Check if blinking reached */
    if(CURRENT_TIME > AWAKE_END) 
    {
      if (BLINK_FLAG == true)                     /* Switch LED off ? */
      {
        SUB_NEO_SEND_COLOR(0, 0, 0);              /* Switch Neopixel LED off */
        BLINK_FLAG = false;                       /* Set blink flag false */
      }
      else                                        /* Switch LED on ? */
      {
        SUB_NEO_SEND_COLOR(128, 0, 128);          /* Set cyan color on Neopixel LED */
        BLINK_FLAG = true;                        /* Set blink flag true */
      }
      /* Set new blink value to long blink delay */
      AWAKE_END = CURRENT_TIME + LONG_BLINK_DELAY;             
    }
  }
  while (true);
} 


/****************************************************************************************/
/****************************************************************************************/
/* HOMEMATIC SUBROUTINES                                                                */
/****************************************************************************************/ 
/****************************************************************************************/

 
/****************************************************************************************/
/* Transmits the byte via the homematic interface                                       */
/* interface                                                                            */
/*      First parameter: Byte value to be transmitted                                   */
/****************************************************************************************/ 
void SUB_TRANS_HOMEMATIC (byte TRANS_BYTE)
{
  /* Transmit D0 */
  if ((TRANS_BYTE & B00000001) == 0)
    digitalWrite(HOMEMATIC_D0, LOW);
  else
    digitalWrite(HOMEMATIC_D0, HIGH);
  /* Transmit D1 */
  if ((TRANS_BYTE & B00000010) == 0)
    digitalWrite(HOMEMATIC_D1, LOW);
  else
    digitalWrite(HOMEMATIC_D1, HIGH);
  /* Transmit D2 */
  if ((TRANS_BYTE & B00000100) == 0)
    digitalWrite(HOMEMATIC_D2, LOW);
  else
    digitalWrite(HOMEMATIC_D2, HIGH);  
  /* Transmit D3 */
  if ((TRANS_BYTE & B00001000) == 0)
    digitalWrite(HOMEMATIC_D3, LOW);
  else
    digitalWrite(HOMEMATIC_D3, HIGH); 
 /* Transmit D4 */
  if ((TRANS_BYTE & B00010000) == 0)
    digitalWrite(HOMEMATIC_D4, LOW);
  else
    digitalWrite(HOMEMATIC_D4, HIGH); 
 /* Transmit D5 */
  if ((TRANS_BYTE & B00100000) == 0)
    digitalWrite(HOMEMATIC_D5, LOW);
  else
    digitalWrite(HOMEMATIC_D5, HIGH); 
 /* Transmit D6 */
  if ((TRANS_BYTE & B01000000) == 0)
    digitalWrite(HOMEMATIC_D6, LOW);
  else
    digitalWrite(HOMEMATIC_D6, HIGH); 
 /* Transmit D7 */
  if ((TRANS_BYTE & B10000000) == 0)
    digitalWrite(HOMEMATIC_D7, LOW);
  else
    digitalWrite(HOMEMATIC_D7, HIGH);

  delay (10);                                     /* 10ms delay to stable data */
  digitalWrite(HOMEMATIC_ACK, LOW);               /* Set data acknowledge low */
  delay (300);                                    /* 300ms delay for acknowledge low delay */
  digitalWrite(HOMEMATIC_ACK, HIGH);              /* Set data acknowledge high */
}


/****************************************************************************************/
/****************************************************************************************/
/* OTHER SUBROUTINES                                                                    */
/****************************************************************************************/ 
/****************************************************************************************/


/****************************************************************************************/
/* Error handler with endless red blinking                                              */
/****************************************************************************************/ 
void SUB_ERROR_ROUTINE (void)
{
  do                                              /* Endless loop */
  {
    SUB_NEO_SEND_COLOR(0, 0, 0);                  /* Switch Neopixel LED off */
    delay (SHORT_BLINK_DELAY);                    /* Blink delay for short blinking */ 
    SUB_NEO_SEND_COLOR(0, 128, 0);                /* Set red color on Neopixel LED */
    delay (SHORT_BLINK_DELAY);                    /* Blink delay for short blinking */       
  }
  while (1);
}
