/****************************************************************************************/
/* Stamp.ino                                                                            */
/* Collection of subroutines to communicate with vic Control stamp                      */
/*                                                                                      */
/* SUB_RECEIVE_ANSWER_VIC   Subroutine to receive answers to vic Control stamp requests */
/* SUB_RECEIVE_MESSAGE_VIC  Subroutine to receive message from vic Control stamp        */
/* SUB_STRING_TO_CRC        Subroutine to calculation from string CRC_BUFFER            */
/****************************************************************************************/


/****************************************************************************************/
/* Receive answer from vic Control stamp after request sent                             */
/* Return value: True: Received correctly, False: Not received correctly                */
/* Receive Mode:                                                                        */
/*         0 = Check on boot complete message: v1MBoot00F88F                            */
/*         1 = Check FW version                                                         */
/*         2 = Check plattform                                                          */
/*         3 = Check language                                                           */
/*         4 = No error message                                                         */
/****************************************************************************************/ 
boolean SUB_RECEIVE_ANSWER_VIC (byte RECEIVE_MODE)
{
  boolean BREAK_LOOP;                             /* Flag for breaking the loop: False = No break, True = break */ 
  char READ_CHAR;                                 /* Read char from serial interface */
  byte READ_LOOP;                                 /* Determine number of chars read from the serial interface */
  uint16_t CRC_FROM_STRING;                       /* CRC extracted from the received string */
  uint16_t CRC_CHECKSUM;                          /* CRC checksum calculated */

  BREAK_LOOP = false;                             /* Set break loop flag true */ 
  READ_LOOP = 0;                                  /* Set read loop counter to zero */
  RECEIVE_BUFFER = "";                            /* Empty receiver buffer */

  /* Loop as long as chars are available or \r detected */
  while ((Serial1.available() > 0) && (BREAK_LOOP == false))               
  {        
    READ_CHAR = Serial1.read();                   /* Read a byte */
    if (READ_CHAR != '\r')                        /* End of message not reached ? */
    {
      if (READ_LOOP < RECEIVE_BUFFER_SIZE)        /* No string overflow ? */
      {
        RECEIVE_BUFFER += READ_CHAR;              /* Add char to string */
        READ_LOOP += 1;                           /* Increase read loop by 1 */
      }                                           
      else                                        /* String overflow ? */
      {
        BREAK_LOOP = true;                        /* Set break loop true */
        #ifdef _DEBUG_
          Serial.println ("    Receiver overflow error!");  
        # endif
        return false;                             /* Return with error */
      }
    }
    else                                          /* End of message reached ? */
    {
      BREAK_LOOP = true;                          /* Set break loop true */
    }
  }
  #ifdef _DEBUG_
    Serial.print ("    Read ");
    Serial.print (READ_LOOP);
    Serial.print (" chars from vicControl stamp with following text: *");
    Serial.print (RECEIVE_BUFFER);
    Serial.println ("*"); 
  # endif

  /* Mode 0: Booting complete message */
  if (RECEIVE_MODE == 0)                          /* Mode 0: Check on boot complete message */
  {
    if (RECEIVE_BUFFER == BOOT_READY)             /* Boot ready message correctly received ? */
    {
      #ifdef _DEBUG_
        Serial.println ("      Boot message received correctly!");  
      # endif      
      return true;                                /* Return without error */
    }
    else                                          /* Boot ready message not correctly received ? */
    {
      #ifdef _DEBUG_
        Serial.println ("      Error: Boot message NOT received correctly!");  
      # endif      
      return false;                               /* Return with error */
    }    
  }

  /* Mode 4: No error message */
  if (RECEIVE_MODE == 4)                          /* Mode 4: Check on no error message */
  {
    if (RECEIVE_BUFFER == NO_ERROR)               /* No error message correctly received ? */   
    {
      #ifdef _DEBUG_
        Serial.println ("      OK received!");  
      # endif      
      return true;                                /* Return without error */
    }  
    else                                          /* No error message not correctly received ? */                                 
    {
      #ifdef _DEBUG_
        Serial.println ("      Error: Wrong OK message received!");  
      # endif      
      return false;                               /* Return with error */
    }    
  }

  /* Mode 1-3: Analyse message */
  /* Step 1: Check header */
  TEMP_BUFFER = RECEIVE_BUFFER.substring(0, 3);   /* Extract first three characters of received string */
  if (TEMP_BUFFER != "v1A")                       /* First three characters are not header v1 and A=answer ? */
  {
    #ifdef _DEBUG_
      Serial.print ("    Error: Wrong header ");  
      Serial.print (TEMP_BUFFER);
      Serial.println (" received!");
    # endif 
    return false;                                 /* Return with error */           
  }
  /* Step 2: Check checksum */
  /* Extract checksum as received */
  TEMP_BUFFER = RECEIVE_BUFFER.substring(READ_LOOP-4, READ_LOOP); 
  /* Generate CRC buffer string */    
  TEMP_BUFFER.toCharArray(CRC_BUFFER, READ_LOOP+1);     
  CRC_FROM_STRING = SUB_STRING_TO_CRC ();         /* Call subroutine to calculate integer value of message string */
  /* Calculate checksum from message */
  TEMP_BUFFER = RECEIVE_BUFFER.substring(0, READ_LOOP-4);
  /* Convert string to char array */
  TEMP_BUFFER.toCharArray(CRC_BUFFER, READ_LOOP+1);
  /* Generate checksum according CRC-CCIT (CRC16): Mode: forward, Start value = 0xFFFF, Polynom = 0x1021 */
  CRC_CHECKSUM = crc16((uint8_t *) CRC_BUFFER, READ_LOOP-4, 0x1021, 0xFFFF, 0x0000, false, false );
  if (CRC_FROM_STRING  != CRC_CHECKSUM)           /* Calculated checksum = Extracted checksum ? */
  {
    #ifdef _DEBUG_    
      Serial.print ("    Error: Wrong checksum, expected: ");
      Serial.print (CRC_FROM_STRING);
      Serial.print (" DEC, ");
      Serial.print (CRC_FROM_STRING, HEX);
      Serial.print (" HEX, calculated: ");
      Serial.print (CRC_CHECKSUM);
      Serial.print (" DEC, ");
      Serial.print (CRC_CHECKSUM, HEX);      
      Serial.println (" HEX");
    # endif
    return false;                                 /* Return with error */           
  }
  /* Step 3: Extract payload */
  TEMP_BUFFER = RECEIVE_BUFFER.substring(3, READ_LOOP-4);
  #ifdef _DEBUG_
    switch (RECEIVE_MODE)                         /* Display right message */
    {
      case 1:
        Serial.print ("      VicControl stamp version: ");
        break;
      case 2:
        Serial.print ("      VicControl stamp plattform: ");
        break;
      case 3:
        Serial.print ("      VicControl stamp language: ");
        break;        
      default:
        break; 
    }            
    Serial.println (TEMP_BUFFER);                 /* Display payload information */
  # endif  
  return true;                                    /* Return with no error */
}


/****************************************************************************************/
/* Receive message from vic Control stamp                                                */
/****************************************************************************************/ 
void SUB_RECEIVE_MESSAGE_VIC (void)
{
  boolean BREAK_LOOP;                             /* Flag for breaking the loop: False = No break, True = break */ 
  char READ_CHAR;                                 /* Read char from serial interface */
  byte READ_LOOP;                                 /* Determine number of chars read from the serial interface */
  
  uint16_t CRC_FROM_STRING;                       /* CRC extracted from the received string */
  uint16_t CRC_CHECKSUM;                          /* CRC checksum calculated */ 
  uint16_t MESSAGE_LENGTH;                        /* Message length of payload information */ 
  uint16_t COMMAND;                               /* Command extracted from payload information */
  
  BREAK_LOOP = false;                             /* Set break loop flag true */ 
  READ_LOOP = 0;                                  /* Set read loop counter to zero */
  RECEIVE_BUFFER = "";                            /* Empty receiver buffer */

  #ifdef _DEBUG_
    Serial.println ("Received message from vicControl stamp!");
  # endif 
  /* Loop as long as chars are available or \r detected */ 
  while ((Serial1.available() > 0) && (BREAK_LOOP == false))               
  {        
    READ_CHAR = Serial1.read();                   /* Read a byte */
    if (READ_CHAR != '\r')                        /* End of message not reached ? */
    {
      if (READ_LOOP < RECEIVE_BUFFER_SIZE)        /* No string overflow ? */
      {
        RECEIVE_BUFFER += READ_CHAR;              /* Add char to string */
        READ_LOOP += 1;                           /* Increase read loop by 1 */
      }                                           
      else                                        /* String overflow ? */
      {
        BREAK_LOOP = true;                        /* Set break loop true */
        #ifdef _DEBUG_
          Serial.println ("    Receiver overflow error!");  
        # endif
        return;                                   /* Return from subroutine */
      }
    }
    else                                          /* End of message reached ? */
    {
      BREAK_LOOP = true;                          /* Set break loop true */
    }
  }
  #ifdef _DEBUG_
    Serial.print ("  Read ");
    Serial.print (READ_LOOP);
    Serial.print (" chars from vicControl stamp with following text: *");
    Serial.print (RECEIVE_BUFFER);
    Serial.println ("*"); 
  # endif 
  
  /* Step 1: Check message header */
  TEMP_BUFFER = RECEIVE_BUFFER.substring(0, 7);   /* Extract the header of received string */
  if (TEMP_BUFFER != "v1MStCh")                   /* First eight characters do not match the message string ? */
  {
    #ifdef _DEBUG_
      Serial.print ("    Error: Wrong header ");  
      Serial.print (TEMP_BUFFER);
      Serial.println (" received!");      
    # endif 
    return;                                       /* Return from sub routine */           
  }

  /* Step 2: Check checksum */
  /* Extract checksum as received */
  TEMP_BUFFER = RECEIVE_BUFFER.substring(READ_LOOP-4, READ_LOOP); 
  /* Generate CRC buffer string */    
  TEMP_BUFFER.toCharArray(CRC_BUFFER, READ_LOOP+1);     
  CRC_FROM_STRING = SUB_STRING_TO_CRC ();         /* Call subroutine to calculate integer value of message string */
  /* Calculate checksum from message */
  TEMP_BUFFER = RECEIVE_BUFFER.substring(0, READ_LOOP-4);
  /* Convert string to char array */
  TEMP_BUFFER.toCharArray(CRC_BUFFER, READ_LOOP+1);
  /* Generate checksum according CRC-CCIT (CRC16): Mode: forward, Start value = 0xFFFF, Polynom = 0x1021 */
  CRC_CHECKSUM = crc16((uint8_t *) CRC_BUFFER, READ_LOOP-4, 0x1021, 0xFFFF, 0x0000, false, false );
  if (CRC_FROM_STRING  != CRC_CHECKSUM)           /* Calculated checksum = Extracted checksum ? */
  {
    #ifdef _DEBUG_    
      Serial.print ("    Error: Wrong checksum, expected: ");
      Serial.print (CRC_FROM_STRING);
      Serial.print (" DEC, ");
      Serial.print (CRC_FROM_STRING, HEX);
      Serial.print (" HEX, calculated: ");
      Serial.print (CRC_CHECKSUM);
      Serial.print (" DEC, ");
      Serial.print (CRC_CHECKSUM, HEX);      
      Serial.println (" HEX");
    # endif
    return;                                       /* Return from sub routine */           
  }
  /* Step 3: Extract payload length */
  TEMP_BUFFER = RECEIVE_BUFFER.substring(7, 9);   /* Extract length information of received string */
  MESSAGE_LENGTH = TEMP_BUFFER.toInt ();          /* Convert string to integer */
  /* Length outside of the range ? */
  if ((MESSAGE_LENGTH < 1) && (MESSAGE_LENGTH > RECEIVE_BUFFER_SIZE))
  {
    #ifdef _DEBUG_
      Serial.print ("    Error: Wrong length ");  
      Serial.print (MESSAGE_LENGTH);
      Serial.println (" received!");     
    # endif     
    return;                                       /* Return from sub routine */ 
  }
  /* Step 4: Extract payload information */ 
  /* Extract payload information */
  TEMP_BUFFER = RECEIVE_BUFFER.substring(9, 9+MESSAGE_LENGTH);   
  #ifdef _DEBUG_
    Serial.print ("    Payload ");  
    Serial.print (TEMP_BUFFER);
    Serial.println (" received!"); 
    SUB_NEO_SEND_COLOR(0, 0, 0);                  /* Switch Neopixel LED off */
    delay (200);                                  /* 200ms blinking delay */
    SUB_NEO_SEND_COLOR(128, 0, 0);                /* Set green color on Neopixel LED */    
  # endif  
  /* Step 5: Check for X commands */
  TEMP_BUFFER = RECEIVE_BUFFER.substring(9, 10);
  if (TEMP_BUFFER != "X")                         /* First eight characters do not match the message string ? */
  {
    return;                                       /* Return from sub routine */           
  }
  /* Step 6: Extract command */
  TEMP_BUFFER = RECEIVE_BUFFER.substring(10, 10+MESSAGE_LENGTH-1);   
  COMMAND = TEMP_BUFFER.toInt ();                 /* Convert string to integer */
  /* Command outside of the range ? */
  if ((COMMAND < 1) && (COMMAND > 255))
  {
    #ifdef _DEBUG_
      Serial.print ("    Error: Not valid command ");  
      Serial.print (COMMAND);
      Serial.println (" received!");     
    # endif     
    return;                                       /* Return from sub routine */ 
  }
  else                                            /* Correct command received */
  {
    SUB_TRANS_HOMEMATIC (COMMAND);                /* Transmit command via Homematic interface */
  #ifdef _DEBUG_
    Serial.print ("  Command ");
    Serial.print (COMMAND);
    Serial.println (" executed!"); 
  # endif     
  }  
}


/****************************************************************************************/
/* CRC calculation from string CRC_BUFFER                                               */
/* Return value: Integer value for CRC buffer                                           */
/****************************************************************************************/ 
uint16_t SUB_STRING_TO_CRC (void)
{
  byte CRC_THOUSAND;                              /* CRC thousand value */
  byte CRC_HUNDRET;                               /* CRC hundret value */
  byte CRC_TEN;                                   /* CRC ten value */
  byte CRC_ONE;                                   /* CRC one value */

  /* Calculate thousand digit */
  CRC_THOUSAND = (CRC_BUFFER[0] <= '9') ? CRC_BUFFER[0] - '0' : CRC_BUFFER[0] - '7';
  /* Calculate hundret digit */
  CRC_HUNDRET =  (CRC_BUFFER[1] <= '9') ? CRC_BUFFER[1] - '0' : CRC_BUFFER[1] - '7'; 
  /* Calculate ten digit */
  CRC_TEN =      (CRC_BUFFER[2] <= '9') ? CRC_BUFFER[2] - '0' : CRC_BUFFER[2] - '7';
  /* Calculate one digit */   
  CRC_ONE =      (CRC_BUFFER[3] <= '9') ? CRC_BUFFER[3] - '0' : CRC_BUFFER[3] - '7';
  /* Return integer value */  
  return ((4096*CRC_THOUSAND) + (256 * CRC_HUNDRET) + (16 * CRC_TEN) + CRC_ONE);
}
