/****************************************************************************************/
/* Interface to vic Control stamp with Homematic integration                            */
/*                                                                                      */
/* Version: 1.0                                                                         */
/* Date: 01.01.2021                                                                     */
/* Compiled with Arduino 1.8.12                                                         */
/* ESP32 arduino extension: 1.0.4                                                       */
/* Board: ESP32 Pico Kit                                                                */
/*                                                                                      */
/* Pin asignments:                                                                      */      
/*     Homematic D0     IO04   Output                                                   */
/*     Homematic D1     IO15   Output                                                   */
/*     Homematic D2     IO13   Output                                                   */
/*     Homematic D3     IO14   Output                                                   */
/*     Homematic D4     IO27   Output                                                   */
/*     Homematic D5     IO33   Output                                                   */
/*     Homematic D6     IO32   Output                                                   */
/*     Homematic D7     IO26   Output                                                   */
/*     Homematic ACK    IO25   Output                                                   */
/*     RXD              IO34   Input                                                    */
/*     TXD              IO19   Output                                                   */
/*     MULTIPLEXER      IO23   Output                                                   */
/*     BUTTON POWER     IO35   Input                                                    */
/*     BUTTON RECOVER   IO38   Input                                                    */
/*     BUTTON BOOT      IO37   Input                                                    */
/*     POWER OFF        IO22   Output                                                   */
/*     RESET VICCONTROL IO18   Output                                                   */
/*     RECOV VICCONTROL IO05   Output                                                   */
/*     NEOPIXEL LED     IO21   Output                                                   */
/*     MUTE             IO00   Output                                                   */
/*                                                                                      */
/* Buttons:                                                                             */
/*     BUTTON_A = Button POWER                                                          */
/*       Short press when power is Off: Switch device on                                */
/*       Short press when power is On:  Toggle between speech recognizer On/Off         */
/*       Long press when power is On: Switch device off                                 */
/*     BUTTON_B = Button RECOVER                                                        */
/*       Short press: Toggle between normal mode and vic Control stamp loading mode     */
/*       Long press: Toggle between normal mode and vic Control stamp recover mode      */
/*     BUTTON_C = Button BOOTLOADER                                                     */
/*       Short press: Reset vic Control stamp                                           */
/*       Long press: Switch into ESP32 bootloader mode                                  */
/*                                                                                      */
/* Libraries:                                                                           */
/*     FastLED for ESP32: https://github.com/FastLED/FastLED                            */
/*     Button 2 library: https://github.com/LennartHennigs/Button2                      */
/*     CRC library: https://github.com/RobTillaart/CRC                                  */
/* Information:                                                                         */
/* Arduino OTA: https://lastminuteengineers.com/esp32-ota-updates-arduino-ide/          */
/****************************************************************************************/

/****************************************************************************************/
/* Libraries                                                                            */
/****************************************************************************************/
#include "FastLED.h"                              /* Include Fast LED NEO Pixel library */
#include "Button2.h"                              /* Include Button 2 library */
#include "CRC16.h"                                /* Include CRC 16 library */
#include "CRC.h"                                  /* Include CRC library */
#include <ArduinoOTA.h>                           /* Include OTA library */


/****************************************************************************************/
/* Definitions of hardware pins                                                         */
/****************************************************************************************/
#define HOMEMATIC_D0          4                   /* Output: Homematic D0 pin */
#define HOMEMATIC_D1         15                   /* Output: Homematic D1 pin */
#define HOMEMATIC_D2         13                   /* Output: Homematic D2 pin */
#define HOMEMATIC_D3         14                   /* Output: Homematic D3 pin */
#define HOMEMATIC_D4         27                   /* Output: Homematic D4 pin */
#define HOMEMATIC_D5         33                   /* Output: Homematic D5 pin */
#define HOMEMATIC_D6         32                   /* Output: Homematic D6 pin */
#define HOMEMATIC_D7         26                   /* Output: Homematic D7 pin */
#define HOMEMATIC_ACK        25                   /* Output: Homematic DU3 data acknowledge pin */
#define MULTIPLEXER          23                   /* Output: Multiplexer 4053: 0 = vic Control stamp by ESP, 1 = vic Control stamp by USB */
#define BUTTON_POWER         35                   /* Input: Power/Mode Button, 0 = Pressed, 1 = Not pressed */
#define BUTTON_RECOVER       38                   /* Input: Load/Recover Button, 0 = Pressed, 1 = Not pressed */
#define BUTTON_BOOT          37                   /* Input: Bootloader Button, 0 = Pressed, 1 = Not pressed */
#define POWER_OFF            22                   /* Output: Power On/Off, 0 = Off, 1 = On */
#define RESET_VICCONTROL     18                   /* Output: Reset vic Control stamp, 0 = Reset, 1 = Normal operation */
#define RECOV_VICCONTROL      5                   /* Output: Recover vic Control stamp, 0 = Recover, 1 = Normal operation */
#define RXD1                 34                   /* RXD for serial interface 1 for vic Control stamp */
#define TXD1                 19                   /* TXD for serial interface 1 for vic Control stamp */
#define MUTE                  0                   /* Output: Mute, 1 = Unmuted, 0 = Muted */


/****************************************************************************************/
/* Definitions for general settings                                                     */
/****************************************************************************************/
/* If this is uncommented here, status messages will be output via the serial interface, if commented out then disabled ! */
// #define _DEBUG_  


/****************************************************************************************/
/* Definitions for FastLED library                                                      */
/****************************************************************************************/
FASTLED_USING_NAMESPACE                           /* Definitions for ESP32 for Fastled */                         
#if defined(FASTLED_VERSION) && (FASTLED_VERSION < 3001000)
  #warning "Requires FastLED 3.1 or later; check github for latest code."
#endif

#define LED_DATA_PIN    21                        /* Led data pin definitions */ 
#define LED_TYPE        WS2812B                   /* Led type definition */
#define LED_COLOR_ORDER RGB                       /* Led color order definition */
#define LED_NUM_LEDS    1                         /* Number of Leds */
#define LED_FRAMES_PER_SECOND  120                /* Constant to define frames per seconds */
CRGB leds[LED_NUM_LEDS];                          /* Reserve buffer for Led data */

#define FASTLED_SHOW_CORE 0                       /* Define the core to run FastLED.show() */
/* Task handles for use in the notifications */
static TaskHandle_t FAST_LED_SHOW_TASK_HANDLE = 0;    
static TaskHandle_t USER_TASK_HANDLE = 0;


/****************************************************************************************/
/* Definitions for Wifi connection for bootloader                                       */
/****************************************************************************************/
IPAddress FIXED_IP(192, 168, 123, 235);           /* Fixed IP address */
IPAddress GATEWAY(192, 168, 123, 1);              /* Gateway address */
IPAddress SUBNET(255, 255, 255, 0);               /* Subnet mask */
const char* WLAN_SSID = "Admin";                  /* SSDI of WLAN */
const char* WLAN_PWD  = "Admin";                  /* Password of WLAN */


/****************************************************************************************/
/* Definitions for Button 2 library                                                     */
/****************************************************************************************/
Button2 BUTTON_A, BUTTON_B, BUTTON_C;             /* Definitions for 3 buttons: Button_A = Power/Mode Button, Button_B = Load/Recover, Button_C = Boot/Reset */


/****************************************************************************************/
/* Definitions for CRC library                                                          */
/****************************************************************************************/
CRC16 crc;                                        /* Definitions for CRC library */


/****************************************************************************************/
/* Constants                                                                            */
/****************************************************************************************/
const char* SW_VERSION = "1.0";                   /* Define FW version */
const byte RECEIVE_BUFFER_SIZE = 160;             /* 160 byte buffer size */
/* Activate speech recognizer */
const String SPEECH_ACTIVATE   = "v1CRecg00C32C\r";
/* Deactivate speech recognizer */
const String SPEECH_DEACTIVATE = "v1CStop00A146\r";
/* Request version of firmware */
const String FIRMWARE_VERSION = "v1CGetV0022B6\r";
/* Request plattform */
const String REQUEST_PLATTFORM = "v1CGtPN009BC0\r";
/* Request language */
const String REQUEST_LANGUAGE = "v1CGLng0003E7\r";
const String PROMPT_1 = "v1CPlPr011BDB3\r";       /* Play prompt 1 = OK standard high tone */
const String PROMPT_2 = "v1CPlPr0128DD0\r";       /* Play Prompt 2 = Back standard low tone */
const String PROMPT_3 = "v1CPlPr0139DF1\r";       /* Play prompt 3 = Triggerword detected */
const String PROMPT_4 = "v1CPlPr014ED16\r";       /* Play prompt 4 = Ready */
const String PROMPT_5 = "v1CPlPr015FD37\r";       /* Play prompt 5 = Standby */
const String PROMPT_6 = "v1CPlPr016CD54\r";       /* Play prompt 6 = Load mode */
const String PROMPT_7 = "v1CPlPr017DD75\r";       /* Play prompt 7 = Recover mode */
const String PROMPT_8 = "v1CPlPr0182C9A\r";       /* Play prompt 8 = Bootloader mode */
const String NO_ERROR = "v1ARetn03000E9C4";       /* No error string */
const String BOOT_READY = "v1MBoot00F88F";        /* Boot ready string */
const uint16_t MESSAGE_RECEIVE_DELAY = 100;       /* Delay to wait from sending command to vic Control stamp until reception of answer */
const uint16_t PLAY_TIME_DELAY = 1000;            /* Delay to wait until message finished playing */
const uint16_t MULTIPLEXER_DELAY = 300;           /* Delay for multiplexer switching */
const uint16_t LONG_BLINK_DELAY = 1000;           /* Delay for long blinking */
const uint16_t SHORT_BLINK_DELAY = 300;           /* Delay for short blinking */


/****************************************************************************************/
/* Variables                                                                            */
/****************************************************************************************/
boolean BUTTON_SHORT_POWER_PRESSED;               /* Flag if power button is short pressed: True: Pressed */
boolean BUTTON_LONG_POWER_PRESSED;                /* Flag if power button is long pressed True: Pressed */
boolean BUTTON_LOAD_PRESSED;                      /* Flag if load button is pressed True: Pressed */
boolean BUTTON_RECOVER_PRESSED;                   /* Flag if recover button is pressed True: Pressed */
boolean BUTTON_BOOTLOADER_PRESSED;                /* Flag if bootloader button is pressed True: Pressed */
boolean BUTTON_RESET_PRESSED;                     /* Flag if reset button is pressed True: Pressed */
boolean STATE_POWER_ON;                           /* Flag for power on state: False: Off, True: On */
boolean STATE_LOAD_ON;                            /* Flag for load state: False: Normal, True: Load state */ 
boolean STATE_RECOVER_ON;                         /* Flag for recover state: False: Normal, True: Recover state */ 
byte LOOP_COUNTER;                                /* General loop counter */    
String RECEIVE_BUFFER = "";                       /* String to receive data */                          
String TEMP_BUFFER = "";                          /* String for temporary handling of data */  
char CRC_BUFFER [RECEIVE_BUFFER_SIZE] = {0};      /* String for CRC_BUFFER */


/****************************************************************************************/
/* Setup procedure                                                                      */
/****************************************************************************************/
void setup() 
{  
  RECEIVE_BUFFER.reserve (RECEIVE_BUFFER_SIZE);   /* Reserve byte buffer for receive buffer */ 
  TEMP_BUFFER.reserve (RECEIVE_BUFFER_SIZE);      /* Reserve byte buffer for temporarry buffer */
    
  /* Step 1: Set power On holding function, muting and show green light */
  pinMode(POWER_OFF, OUTPUT);                     /* Configure Power on/off as output */
  digitalWrite(POWER_OFF, HIGH);                  /* Set Power on/off to high = On */
  pinMode(MUTE, OUTPUT);                          /* Configure Mute as output */
  digitalWrite(MUTE, LOW);                        /* Set Mute to low = Muted */  
  /* Initialize FastLED with the right configuration for the strip */
  FastLED.addLeds<LED_TYPE,LED_DATA_PIN,LED_COLOR_ORDER>(leds, LED_NUM_LEDS).setCorrection(TypicalLEDStrip);
  /* Create the FastLED show task */
  xTaskCreatePinnedToCore(SUB_FAST_LED_SHOW_TASK, "SUB_FAST_LED_SHOW_TASK", 2048, NULL, 2, &FAST_LED_SHOW_TASK_HANDLE, FASTLED_SHOW_CORE);
  FastLED.delay(1000/LED_FRAMES_PER_SECOND);      /* Set the frame update delay */
  SUB_NEO_SEND_COLOR(128, 0, 0);                  /* Set green color on Neopixel LED */
    
  /* Step 2: Init serial interface for debug information */
  Serial.begin(115200);                           /* Start serial interface for terminal communication at 115.000 Baud */
  /* Display welcome message with initialisation information */
  Serial.println("\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n");
  Serial.println ("*********************************************************");
  Serial.println ("* vicControl stamp interface for Homematic integration  *");
  Serial.println ("*                                                       *");
  Serial.print   ("* V");
  Serial.print   (SW_VERSION);
  Serial.println ("                                      @PT 2022    *");   
  Serial.println ("*********************************************************\n"); 

  #ifdef _DEBUG_
    Serial.println("\n\n\n"); 
    Serial.println ("Debug is enabled...");
    Serial.println(); 
  #else
    Serial.println("\n\n\n"); 
    Serial.println ("Debug is disabled...");
    Serial.println ("Remove commented out line #define _DEBUG_ to enable!");
    Serial.println();     
  # endif  

  #ifdef _DEBUG_
    Serial.println("Booting up device ...");
  # endif   

  /* Step 3: Prepare homematic interface */
  #ifdef _DEBUG_
    Serial.println("  Step 1: Initializing homematic interface: Done");
  # endif   
  digitalWrite(HOMEMATIC_ACK, HIGH);              /* Set Homematic data acknowlwdge to high */
  pinMode(HOMEMATIC_ACK, OUTPUT);                 /* Define Homematic data acknowledge as output */
  digitalWrite(HOMEMATIC_ACK, HIGH);              /* Set data acknowlwdge to high */
  pinMode(HOMEMATIC_D0, OUTPUT);                  /* Define Homematic D0 as output */
  digitalWrite(HOMEMATIC_D0, HIGH);               /* Set Homematic D0 to high */
  pinMode(HOMEMATIC_D1, OUTPUT);                  /* Define Homematic D1 as output */
  digitalWrite(HOMEMATIC_D1, HIGH);               /* Set Homematic D1 to high */
  pinMode(HOMEMATIC_D2, OUTPUT);                  /* Define Homematic D2 as output */
  digitalWrite(HOMEMATIC_D2, HIGH);               /* Set Homematic D2 to high */
  pinMode(HOMEMATIC_D3, OUTPUT);                  /* Define Homematic D3 as output */
  digitalWrite(HOMEMATIC_D3, HIGH);               /* Set Homematic D3 to high */
  pinMode(HOMEMATIC_D4, OUTPUT);                  /* Define Homematic D4 as output */
  digitalWrite(HOMEMATIC_D4, HIGH);               /* Set Homematic D4 to high */
  pinMode(HOMEMATIC_D5, OUTPUT);                  /* Define Homematic D5 as output */
  digitalWrite(HOMEMATIC_D5, HIGH);               /* Set Homematic D5 to high */
  pinMode(HOMEMATIC_D6, OUTPUT);                  /* Define Homematic D6 as output */
  digitalWrite(HOMEMATIC_D6, HIGH);               /* Set Homematic D6 to high */
  pinMode(HOMEMATIC_D7, OUTPUT);                  /* Define Homematic D7 as output */
  digitalWrite(HOMEMATIC_D7, HIGH);               /* Set Homematic D7 to high */  

  /* Step 4: Prepare various other pins */
  #ifdef _DEBUG_
    Serial.println("  Step 2: Initializing pins: Done");
  # endif   
  pinMode(MULTIPLEXER, OUTPUT);                   /* Define USB multiplexer as output */
  digitalWrite(MULTIPLEXER, LOW);                 /* Set vic Control stamp by ESP */
  digitalWrite(RECOV_VICCONTROL, HIGH);           /* Preconfigure Recover vic Control stamp high */
  pinMode(RECOV_VICCONTROL, OUTPUT);              /* Define Recover vic Control stamp as output */
  digitalWrite(RECOV_VICCONTROL, HIGH);           /* No recover mode for vic Control stamp */
  digitalWrite(RESET_VICCONTROL, HIGH);           /* Preconfigure Reset vic Control stamp high */
  pinMode(RESET_VICCONTROL, OUTPUT);              /* Define Reset vic Control stamp as output */
  digitalWrite(RESET_VICCONTROL, HIGH);           /* No reset mode vic Control stamp */
  pinMode(BUTTON_POWER, INPUT);                   /* Define power/mode button as input */  
  pinMode(BUTTON_RECOVER, INPUT);                 /* Define load/recover button as input */ 
  pinMode(BUTTON_BOOT, INPUT);                    /* Define bootloader/reset button as input */    

  /* Step 5: Initialize serial interface for vic Control stamp */
  #ifdef _DEBUG_
    Serial.println ("  Step 3: Init serial interface for vicControl stamp: Done");
  # endif
  /* Init interface to vic Control stamp */
  Serial1.begin(115200, SERIAL_8N1, RXD1, TXD1);  /* Set the data rate for the serial port 1 to 115.000 baud, 8 bit, non parity, one stop bit */
  /* Empty receiver buffer */
  while (Serial1.available())                     /* Not empty receiver buffer? */
  {
    Serial1.read();                               /* Read buffer */
  }
  delay (LONG_BLINK_DELAY);                       /* Blink delay for LED long blinking */
  WiFi.mode(WIFI_OFF);                            /* Switch WIFI off */
  btStop();                                       /* Switch Bluetooth off */

  /* Step 6: Reset vic Control stamp */
  #ifdef _DEBUG_
    Serial.print ("  Step 4: Reset vicControl stamp: ");
  # endif
  digitalWrite(RESET_VICCONTROL, LOW);            /* Reset vic Control stamp */                 
  SUB_NEO_SEND_COLOR(0, 0, 0);                    /* Switch Neopixel LED off */
  delay (LONG_BLINK_DELAY);                       /* Blink delay for LED long blinking */
  digitalWrite(RESET_VICCONTROL, HIGH);           /* Remove Reset vic Control stamp */ 
  SUB_NEO_SEND_COLOR(128, 0, 0);                  /* Set green color on Neopixel LED */
  delay (LONG_BLINK_DELAY);                       /* Blink delay for LED long blinking */
  #ifdef _DEBUG_
    Serial.println ("Done");
  # endif
  
  /* Step 7: Check for presence of vic Control stamp after Reset */
  #ifdef _DEBUG_
    Serial.println ("  Step 5: Waiting for boot complete...");
  # endif  
  LOOP_COUNTER = 0;                               /* Reset the loop counter */
  do                                              /* Repeat until boot message received or timeout */
  {
    SUB_NEO_SEND_COLOR(0, 0, 0);                  /* Switch Neopixel LED off */
    delay (LONG_BLINK_DELAY);                     /* Blink delay for LED long blinking */
    SUB_NEO_SEND_COLOR(128, 0, 0);                /* Set green color on Neopixel LED */
    delay (LONG_BLINK_DELAY);                     /* Blink delay for LED long blinking */
    if (Serial1.available() > 0)                  /* Received something from vic Control stamp ? */
    {
      if (SUB_RECEIVE_ANSWER_VIC (0)== true)      /* Receipt of boot message (Mode 0) OK ? */
      {
        LOOP_COUNTER = 255;                       /* Set for break loop */         
      }
      else                                        /* Receipt of boot message (Mode 0) not OK ? */
      {
        SUB_ERROR_ROUTINE ();                     /* Call sub routine for error handling in endless red blinking loop */
      }
    }
    else                                          /* Nothing received so far ? */
    {
      LOOP_COUNTER += 1;                          /* Increase loop counter */    
    }
  }
  while (LOOP_COUNTER < 10);                      /* Wait for maximum of 20 seconds */

  if (LOOP_COUNTER == 10)                         /* Timeout */
  {
    #ifdef _DEBUG_
      Serial.println ("    Error: Boot timeout!");
    # endif  
    SUB_ERROR_ROUTINE ();                         /* Call sub routine for error handling in endless red blinking loop */       
  }

  /* Step 8: Initialize button library, buttons and variables */
  #ifdef _DEBUG_
    Serial.println ("  Step 6: Set variables and initiate buttons: Done");
  # endif
  BUTTON_SHORT_POWER_PRESSED = false;             /* Set short power button pressed false */
  BUTTON_LONG_POWER_PRESSED = false;              /* Set long power button pressed false */
  BUTTON_LOAD_PRESSED = false;                    /* Set load button pressed false */
  BUTTON_RECOVER_PRESSED = false;                 /* Set recover button pressed false */
  BUTTON_BOOTLOADER_PRESSED = false;              /* Set bootloader button pressed false */
  BUTTON_RESET_PRESSED = false;                   /* Set reset button pressed false */
  STATE_POWER_ON = false;                         /* Set power state to Standby */
  STATE_LOAD_ON = false;                          /* Set load state to Normal */
  STATE_RECOVER_ON = false;                       /* Set recover state to Normal */
  
  BUTTON_A.begin(BUTTON_POWER);                   /* Definitions for power/mode button */
  BUTTON_A.setLongClickHandler(SUB_LONG_PRESS);   /* Attach interrupt handler for power/mode button */
  BUTTON_B.begin(BUTTON_RECOVER);                 /* Definitions for load/recover button */
  BUTTON_B.setLongClickHandler(SUB_LONG_PRESS);   /* Attach interrupt handler for load/recover button */
  BUTTON_C.begin(BUTTON_BOOT);                    /* Definitions for boot/reset button */
  BUTTON_C.setLongClickHandler(SUB_LONG_PRESS);   /* Attach interrupt handler for boot/reset button */

  /* Step 9: Determine FW Version of vic Control stamp */
  #ifdef _DEBUG_
    Serial.println ("  Step 7: Determine FW Version of vicControl stamp...");
  # endif
  Serial1.print (FIRMWARE_VERSION);               /* Request FW version from vic Control stamp */ 
  delay (MESSAGE_RECEIVE_DELAY);                  /* Delay to receive answer from vic Control stamp */
  if (SUB_RECEIVE_ANSWER_VIC (1)== false)         /* Receipt of FW version (Mode = 1) not OK ? */
  {
    SUB_ERROR_ROUTINE ();                         /* Call sub routine for error handling in endless red blinking loop */
  }

  /* Step 10: Request plattform of vic Control stamp */
  #ifdef _DEBUG_
    Serial.println ("  Step 8: Determine plattform of vicControl stamp...");
  # endif
  Serial1.print (REQUEST_PLATTFORM);              /* Request plattform from vic Control stamp */ 
  delay (MESSAGE_RECEIVE_DELAY);                  /* Delay to receive answer from vic Control stamp */
  if (SUB_RECEIVE_ANSWER_VIC (2)== false)         /* Receipt of plattform message (Mode = 2) not OK ? */
  {
    SUB_ERROR_ROUTINE ();                         /* Call sub routine for error handling in endless red blinking loop */
  }

  /* Step 11: Request language of vic Control stamp */
  #ifdef _DEBUG_
    Serial.println ("  Step 9: Determine language of vicControl stamp...");
  # endif
  Serial1.print (REQUEST_LANGUAGE);               /* Request plattform from vic Control stamp */ 
  delay (MESSAGE_RECEIVE_DELAY);                  /* Delay to receive answer from vic Control stamp */
  if (SUB_RECEIVE_ANSWER_VIC (3)== false)         /* Receipt of plattform message (Mode = 3) not OK ? */
  {
    SUB_ERROR_ROUTINE ();                         /* Call sub routine for error handling in endless red blinking loop */
  }

  #ifdef _DEBUG_                                  /* Display boot finished message */
    Serial.println(); 
    Serial.println ("Booting completed successful!");
    Serial.println ("  Playing ready message...");
  # endif
  Serial1.print (PROMPT_4);                       /* Play ready prompt */ 
  delay (MESSAGE_RECEIVE_DELAY);                  /* Delay to receive answer from vic Control stamp */
  if (SUB_RECEIVE_ANSWER_VIC (4)== false)         /* Receipt of no error (Mode = 4) not OK ? */
  {
    SUB_ERROR_ROUTINE ();                         /* Call sub routine for error handling in endless red blinking loop */
  }
  #ifdef _DEBUG_    
    Serial.println ("Ready to speak ...");
    Serial.println(); 
  # endif
  SUB_NEO_SEND_COLOR(128, 0, 0);                  /* Set green color on Neopixel LED */       
  STATE_POWER_ON = true;                          /* Set power state to On */ 
  digitalWrite(MUTE, HIGH);                       /* Set Mute to high = Unmuted */
}


/****************************************************************************************/
/* Main loop                                                                            */
/****************************************************************************************/
void loop()
{
  BUTTON_A.loop();                                /* Call buttom handler for power/mode button */
  BUTTON_B.loop();                                /* Call buttom handler for load/recover button */
  BUTTON_C.loop();                                /* Call buttom handler for boot/reset button */

  /* Step A: Check if any message is pending */
  if (Serial1.available() > 0)                    /* Received something from vic Control ? */
  {
    SUB_RECEIVE_MESSAGE_VIC ();                   /* Call subroutine to receive message */
    #ifdef _DEBUG_   
      Serial.println (); 
    # endif
  }

  /* Step B: Check for power/mode button short pressed = Change mode */
  if (BUTTON_SHORT_POWER_PRESSED == true)         /* Power button short press detected ? */
  {
    SUB_TOGGLE_MODE ();                           /* Call subroutine to activate/deactivate speech recognizer */
  }

  /* Step C: Check for power/mode button long pressed = Power off */
  if (BUTTON_LONG_POWER_PRESSED == true)          /* Power button long press detected ? */
  {
    SUB_NEO_SEND_COLOR(0, 0, 0);                  /* Switch Neopixel LED off */
    digitalWrite(POWER_OFF, LOW);                 /* Set Power on/off to low = Off */  
    do                                            /* Endless loop */
    {}
    while (1);    
  }

  /* Step D: Check for load/recover short pressed = Toggle load menu */
  if (BUTTON_LOAD_PRESSED == true)                /* Load button press detected ? */
  {
    if (STATE_RECOVER_ON == false)                /* Normal mode ? */
    {
      SUB_TOGGLE_LOAD ();                         /* Call subroutine to activate/deactivate load mode */
    }
    else                                          /* Recover mode ? */
    {
      BUTTON_LOAD_PRESSED = false;                /* Reset the load button pressed flag */  
      #ifdef _DEBUG_                             
        Serial.println ("  Error: Locked in recovery mode!");
        Serial.println(); 
      # endif
    }
  }

  /* Step E: Check for load/recover long pressed = Toggle restore menu */
  if (BUTTON_RECOVER_PRESSED == true)             /* Recover button press detected ? */
  {
    if (STATE_LOAD_ON == false)                   /* Normal mode ? */
    {
      SUB_TOGGLE_RECOVER ();                      /* Call subroutine to activate/deactivate recover mode */
    }
    else                                          /* Load mode ? */
    {
      BUTTON_RECOVER_PRESSED = false;             /* Reset the recover button pressed flag */
      #ifdef _DEBUG_                               
        Serial.println ("  Error: Locked in load mode!");
        Serial.println(); 
      # endif      
    }
  }

  /* Step F: Check for bootloader/reset short pressed = Reset menu */
  if (BUTTON_RESET_PRESSED == true)               /* Reset button press detected ? */  
  {
    SUB_RESET_VIC ();                             /* Call subroutine to reset vic Control stamp */                   
  }

  /* Step G: Check for bootloader/reset long pressed = Bootloader menu */
  if (BUTTON_BOOTLOADER_PRESSED == true)          /* Bootloader button press detected ? */
  {
    SUB_ESP_BOOTLOADER ();                        /* Call subroutine to reset vic Control stamp */
    BUTTON_BOOTLOADER_PRESSED = false;            /* Reset the boatloader button pressed flag */    
  }
}
